from logging import getLogger
from os import getenv
from typing import Any, Callable, Iterator, List, Optional, Text, Union

from httpx import HTTPError, Timeout
from typefit import api
from typefit import httpx_models as hm

from .models import *
from .types import UndefinedSerializer

logger = getLogger("douze.api")


class DoApi(api.SyncClient):
    BASE_URL = "https://api.digitalocean.com/v2/"

    def __init__(self, api_token: Text = getenv("DO_API_TOKEN", "")):
        super().__init__()
        self.api_token = api_token
        self.helper.http.timeout = Timeout(30.0)

    def init_serialize(self) -> Callable[[Any], Any]:
        """
        We use a custom serializer to avoid sending some fields that are
        None while they should not
        """

        return UndefinedSerializer().serialize

    def headers(self) -> Optional[hm.HeaderTypes]:
        """
        The API expects the token in the headers
        """

        return {"Authorization": f"Bearer {self.api_token}"}

    def extract(self, data: Any, hint: Any) -> Any:
        """
        DigitalOcean had this fantastic idea that every API response is going
        to be a dictionary and the content you're looking for is always going
        to be in a different key that's more or less logic.

        The hint here serves to give the name of that key. By example, if the
        hint is "droplet" the response is expected to be something like

        >>> {
        >>>     "droplet": {
        >>>         # actual droplet content
        >>>     }
        >>> }

        Given that situation, the actual droplet content will be extracted and
        returned.

        If no hint is given, data is returned as-is.

        Parameters
        ----------
        data
            Data to exctract from
        hint
            Name of the key to look for
        """

        if hint:
            return data[hint]

        return data

    def decode(self, resp: hm.Response, hint: Any) -> Any:
        """
        Some responses will be 204 without any content, in that case don't
        try anything. But most of the time it's JSON.
        """

        if resp.status_code != 204:
            return resp.json()

    def raise_errors(self, resp: hm.Response, hint: Any) -> None:
        """
        Hacky way to report API errors to the developer
        """

        try:
            super().raise_errors(resp, hint)
        except HTTPError:
            # noinspection PyBroadException
            try:
                logger.error("API error: %s", resp.json())
            except Exception:
                pass

            raise

    def _iterate_collection(self, page_getter, key, **kwargs):
        """
        General system to iterate through all the pages of a given collection.
        It will return an iterator that goes through all items in all the pages
        of the collection.

        Parameters
        ----------
        page_getter
            Function that will get a specific page. It must accept "page" as
            argument and return a Collection.
        key
            Look for the data into that key inside the collection (like if it's
            the droplets collection, the key will be "droplets").
        kwargs
            Extra kwargs to be passed to the page_getter when called (on top
            of "page").
        """

        page = 1
        found = 0
        count = 1

        while found < count and page < 1000:
            collection = page_getter(page=page, **kwargs)
            items = getattr(collection, key)
            yield from items

            if collection.meta:
                count = collection.meta.total
                found += len(items)
                page += 1
            else:
                break

    @api.get("databases?page={page}")
    def _db_cluster_list(self, page) -> DatabaseClusterCollection:
        """
        Fetches one page of database clusters list
        """

    def db_cluster_list(self) -> Iterator[DatabaseCluster]:
        """
        Lists all database clusters present on the account

        Notes
        -----
        Since new versions of DB engines get added and so forth, we don't want
        to avoid a crash in case things are not fit. That's why the databases
        attributes also allows to not fit. Here we're checking that we're
        returning only the instances that are indeed recognized. The other ones
        are just ignored.
        """

        for cluster in self._iterate_collection(self._db_cluster_list, "databases"):
            if isinstance(cluster, DatabaseCluster):
                yield cluster

    @api.post("databases", json=lambda cluster: cluster, hint="database")
    def db_cluster_create(self, cluster: DatabaseClusterCreate) -> DatabaseCluster:
        """
        Creates a database cluster
        """

    @api.get("databases/{cluster_id}", hint="database")
    def db_cluster_get(self, cluster_id: Text) -> DatabaseCluster:
        """
        Returns a database cluster
        """

    @api.get("databases/{cluster_id}/dbs?page={page}")
    def _db_database_list(self, page: int, cluster_id: Text) -> DatabaseCollection:
        """
        Retrieves a single page of the databases list for a cluster
        """

    def db_database_list(self, cluster_id: Text) -> Iterator[Database]:
        """
        Iterates through all the databases present in a cluster
        """

        yield from self._iterate_collection(
            self._db_database_list, "dbs", cluster_id=cluster_id
        )

    @api.post("databases/{cluster_id}/dbs", hint="db", json=lambda database: database)
    def db_database_create(self, cluster_id: Text, database: Database) -> Database:
        """
        Creates a database within a cluster
        """

    @api.delete("databases/{cluster_id}/dbs/{database_name}")
    def db_database_delete(self, cluster_id: Text, database_name: Text) -> None:
        """
        Deletes the specified database within the cluster
        """

    @api.get("databases/{cluster_id}/firewall", hint="rules")
    def db_firewall_list(self, cluster_id: Text) -> List[DatabaseFirewallRule]:
        """
        Lists all firewall rules for a given cluster
        """

    @api.put("databases/{cluster_id}/firewall", json=lambda rules: {"rules": rules})
    def db_firewall_update(
        self,
        cluster_id: Text,
        rules: List[Union[DatabaseFirewallRule, DatabaseFirewallRuleCreate]],
    ) -> None:
        """
        Updates the firewall rules for that cluster
        """

    @api.get("databases/{cluster_id}/users", hint="users")
    def db_user_list(self, cluster_id: Text) -> List[DatabaseUser]:
        """
        Retrieves the list of users found in a cluster
        """

    @api.post("databases/{cluster_id}/users", hint="user", json=lambda user: user)
    def db_user_create(
        self, cluster_id: Text, user: DatabaseUserCreate
    ) -> DatabaseUser:
        """
        Creates a user within a cluster
        """

    @api.delete("databases/{cluster_id}/users/{user_name}")
    def db_user_delete(self, cluster_id: Text, user_name: Text) -> None:
        """
        Deletes a user within a cluster
        """

    @api.get("databases/{cluster_id}/pools", hint="pools")
    def db_pool_list(self, cluster_id: Text) -> List[DatabaseConnectionPool]:
        """
        Lists all connection pools present in the cluster
        """

    @api.post("databases/{cluster_id}/pools", hint="pool", json=lambda pool: pool)
    def db_pool_create(
        self, cluster_id: Text, pool: DatabaseConnectionPoolCreate
    ) -> DatabaseConnectionPool:
        """
        Creates a connection pool for that cluster
        """

    @api.delete("databases/{cluster_id}/pools/{pool_name}")
    def db_pool_delete(self, cluster_id: Text, pool_name: Text) -> None:
        """
        Deletes a connection pool for that cluster
        """

    @api.get("database/{cluster_id}/pools/{pool_name}", hint="pool")
    def db_pool_get(self, cluster_id: Text, pool_name: Text) -> DatabaseConnectionPool:
        """
        Returns a connection pool from that cluster
        """

    @api.get("droplets?page={page}")
    def _droplet_list(self, page: int) -> DropletCollection:
        """
        Lists droplets on a specific page of the list
        """

    def droplet_list(self) -> Iterator[Droplet]:
        """
        Iterates through all the droplets in the account
        """

        yield from self._iterate_collection(self._droplet_list, "droplets")



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, List, Optional, Text, Union
from urllib.parse import quote

from typefit.narrows import DateTime

from .types import Uuid

__all__ = [
    "EntryState",
    "DatabaseSize",
    "DatabaseEngine",
    "PostgreSqlVersion",
    "MySqlVersion",
    "RedisVersion",
    "DatabaseUserRole",
    "MySqlAuthPlugin",
    "DatabaseStatus",
    "DropletStatus",
    "Day",
    "PgBouncerMode",
    "DbFirewallRuleType",
    "DatabaseCluster",
    "DatabaseClusterCreate",
    "DatabaseMaintenanceWindow",
    "DatabaseConnection",
    "DatabaseUser",
    "DatabaseUserCreate",
    "MySqlSettings",
    "Database",
    "DatabaseConnectionPoolCreate",
    "DatabaseConnectionPool",
    "DatabaseFirewallRule",
    "DatabaseFirewallRuleCreate",
    "Droplet",
    "Meta",
    "DatabaseClusterCollection",
    "DatabaseCollection",
    "DropletCollection",
]


class EntryState(Enum):
    present = "present"
    absent = "absent"


class DatabaseSize(Enum):
    db_s_1vcpu_1gb = "db-s-1vcpu-1gb"
    db_s_1vcpu_2gb = "db-s-1vcpu-2gb"
    db_s_2vcpu_4gb = "db-s-2vcpu-4gb"
    db_s_4vcpu_8gb = "db-s-4vcpu-8gb"
    db_s_6vcpu_16gb = "db-s-6vcpu-16gb"
    db_s_8vcpu_32gb = "db-s-8vcpu-32gb"
    db_s_16vcpu_64gb = "db-s-16vcpu-64gb"


class DatabaseEngine(Enum):
    pg = "pg"
    mysql = "mysql"
    redis = "redis"


class PostgreSqlVersion(Enum):
    v10 = "10"
    v11 = "11"
    v12 = "12"
    v13 = "13"
    v14 = "14"
    v15 = "15"
    v16 = "16"


class MySqlVersion(Enum):
    v8 = "8"


class RedisVersion(Enum):
    v5 = "5"
    v6 = "6"
    v7 = "7"


class DatabaseUserRole(Enum):
    primary = "primary"
    normal = "normal"


class MySqlAuthPlugin(Enum):
    mysql_native = "mysql_native_password"
    caching_sha2 = "caching_sha2_password"


class DatabaseStatus(Enum):
    creating = "creating"
    online = "online"
    resizing = "resizing"
    migrating = "migrating"


class Day(Enum):
    monday = "monday"
    tuesday = "tuesday"
    wednesday = "wednesday"
    thursday = "thursday"
    friday = "friday"
    saturday = "saturday"
    sunday = "sunday"


class DbFirewallRuleType(Enum):
    droplet = "droplet"
    k8s = "k8s"
    ip_addr = "ip_addr"
    tag = "tag"


class DropletStatus(Enum):
    new = "new"
    active = "active"
    off = "off"
    archive = "archive"


class PgBouncerMode(Enum):
    session = "session"
    transaction = "transaction"
    statement = "statement"


@dataclass
class DatabaseCluster:
    id: Text
    name: Text
    engine: Union[DatabaseEngine, Text]
    version: Union[PostgreSqlVersion, MySqlVersion, RedisVersion, Text]
    connection: "DatabaseConnection"
    private_connection: "DatabaseConnection"
    num_nodes: int
    size: Union[DatabaseSize, Text]
    region: Text
    status: DatabaseStatus
    created_at: DateTime
    private_network_uuid: Uuid
    users: Optional[List["DatabaseUser"]] = None
    db_names: Optional[List[Text]] = None
    tags: Optional[List[Text]] = None
    maintenance_window: Optional["DatabaseMaintenanceWindow"] = None


@dataclass
class DatabaseClusterCreate:
    name: Text
    engine: Union[DatabaseEngine, Text]
    version: Union[PostgreSqlVersion, MySqlVersion, RedisVersion, Text]
    size: Union[DatabaseSize, Text]
    region: Text
    num_nodes: int
    tags: List[Text] = field(default_factory=list)
    private_network_uuid: Optional[Uuid] = None


@dataclass
class DatabaseMaintenanceWindow:
    day: Day
    hour: Text
    pending: bool
    description: List[Text] = field(default_factory=list)


@dataclass
class DatabaseConnection:
    database: Text
    host: Text
    port: int
    user: Text
    password: Text
    ssl: bool

    @property
    def uri(self) -> Text:
        return (
            f"postgresql://{quote(self.user)}:{quote(self.password)}"
            f"@{quote(self.host)}:{quote(str(self.port))}"
            f"/{quote(self.database)}"
            f'{"?sslmode=require" if self.ssl else ""}'
        )

    @property
    def pg_env(self):
        return {
            "PGPASSWORD": self.password,
            **({"PGSSLMODE": "require"} if self.ssl else {}),
        }

    def pg_flags(self, database: Text = ""):
        if not database:
            database = self.database

        return [
            f"{x}"
            for x in [
                "-U",
                self.user,
                "-h",
                self.host,
                "-p",
                self.port,
                "-d",
                database,
            ]
        ]


@dataclass
class DatabaseUserCreate:
    name: Text
    mysql_settings: Optional["MySqlSettings"] = None


@dataclass
class DatabaseUser(DatabaseUserCreate):
    name: Text = ""
    role: DatabaseUserRole = DatabaseUserRole.normal
    password: Text = ""


@dataclass
class MySqlSettings:
    auth_plugin: MySqlAuthPlugin


@dataclass
class Database:
    name: Text


@dataclass(frozen=True)
class DatabaseFirewallRuleCreate:
    type: DbFirewallRuleType
    value: Text


@dataclass(frozen=True)
class DatabaseFirewallRule(DatabaseFirewallRuleCreate):
    uuid: Text
    created_at: DateTime
    cluster_uuid: Text


@dataclass
class DatabaseConnectionPoolCreate:
    name: Text
    mode: PgBouncerMode
    size: int
    db: Text
    user: Text


@dataclass
class DatabaseConnectionPool(DatabaseConnectionPoolCreate):
    connection: DatabaseConnection
    private_connection: DatabaseConnection


@dataclass
class Droplet:
    id: int
    name: Text
    memory: int
    vcpus: int
    disk: int
    locked: bool
    created_at: DateTime
    status: DropletStatus


@dataclass
class Meta:
    total: int


@dataclass
class Collection:
    meta: Meta = None


@dataclass
class DatabaseClusterCollection(Collection):
    databases: List[Union[DatabaseCluster, Any]] = field(default_factory=list)


@dataclass
class DatabaseCollection(Collection):
    dbs: List[Database] = field(default_factory=list)


@dataclass
class DropletCollection(Collection):
    droplets: List[Droplet] = field(default_factory=list)



import subprocess
from contextlib import contextmanager
from dataclasses import replace
from os import environ
from tempfile import NamedTemporaryFile
from time import sleep, time
from typing import Any, List, NamedTuple, Optional, Sequence, Text

from httpx import Client

from .api import DoApi
from .models import *
from .types import Uuid


class IdemApiError(Exception):
    pass


class Outcome(NamedTuple):
    changed: bool
    output: Optional[Any] = None


class DoIdemApi:
    """
    DigitalOcean idempotent API. It's made to be used as a backend for Ansible
    modules. Basically, instead of describing actions, you describe what you
    would like to see and the API will make sure that it exists.

    Examples
    --------
    You could want to create a database within an existing cluster:

    >>> i = DoIdemApi(DoApi())
    >>> i.psql_database('some-cluster', 'my-db')

    If the database does not exist yet then it will be created while if it
    already exists it will stay in place.
    """

    DEFAULT_PSQL_VERSION = PostgreSqlVersion.v11
    PROVISION_TIMEOUT = 60 * 30
    PROVISION_POLL = 5

    def __init__(self, api: DoApi):
        self.api: DoApi = api
        self._cluster_cache = {}

    def _find_cluster_by_name(self, name) -> Optional[DatabaseCluster]:
        """
        For some reason, the API asks that you set a unique name to clusters
        but then references them based on a different ID. In order to make this
        class workable, it uses cluster names everywhere so this helper is
        used internally to find the clusters from their name and thus their ID.

        Please note that clusters associated to given names are kept in cache,
        so if you're developing new functions and they operate on clusters,
        don't forget to update the cache otherwise you're going to get into
        trouble.

        Parameters
        ----------
        name
            Name of the desired cluster
        """

        if name not in self._cluster_cache:
            cluster = None

            for candidate in self.api.db_cluster_list():
                if candidate.name == name:
                    cluster = candidate
                    break

            self._cluster_cache[name] = cluster

        return self._cluster_cache[name]

    def _find_database_by_name(
        self, cluster_id: Text, name: Text
    ) -> Optional[Database]:
        """
        Finds a specific database by name. Same thing as with the clusters.

        Parameters
        ----------
        cluster_id
            ID of the cluster containing a database.
        name
            Name of the database you're looking for
        """

        for candidate in self.api.db_database_list(cluster_id):
            if candidate.name == name:
                return candidate

    def _get_public_ipv4(self) -> Text:
        """
        Tries to figure out your public IPv4, this is helpful to gain temporary
        access to the firewall in order to copy the databases by example.
        """

        with Client() as c:
            return c.get("https://httpbin.org/get").json()["origin"]

    @contextmanager
    def _allow_self_access(self, cluster_name: Text):
        """
        That's a context manager which will make sure that your IP address is
        allowed in the cluster's firewall during the context.

        If the IP was already allowed, it'll stay there afterwards while if it
        was not there it'll get cleaned up when all is done (or failed).

        Notes
        -----
        The public IP is determined automatically by `_get_public_ipv4()`.

        Parameters
        ----------
        cluster_name
            Name of the cluster to allow the current IP for
        """

        ip = self._get_public_ipv4()
        roll_back, _ = self.db_firewall_rule(
            cluster_name,
            [DatabaseFirewallRuleCreate(DbFirewallRuleType.ip_addr, ip)],
            EntryState.present,
        )

        try:
            yield
        finally:
            if roll_back:
                self.db_firewall_rule(
                    cluster_name,
                    [DatabaseFirewallRuleCreate(DbFirewallRuleType.ip_addr, ip)],
                    EntryState.absent,
                )

    def db_firewall_allow_self(self, cluster_name: Text):
        """
        Ensures that there is a permanent rule allowing this computer's public
        IP address to connect the cluster.

        Parameters
        ----------
        cluster_name
            Name of the cluster to whitelist the IP in
        """

        ip = self._get_public_ipv4()
        return self.db_firewall_rule(
            cluster_name,
            [DatabaseFirewallRuleCreate(DbFirewallRuleType.ip_addr, ip)],
            EntryState.present,
        )

    def db_firewall_rule(
        self,
        cluster_name: Text,
        rules: List[DatabaseFirewallRuleCreate],
        state: EntryState = EntryState.present,
    ) -> Outcome:
        """
        Ensures that this list of rules is either present or absent from the
        cluster's firewall.

        Parameters
        ----------
        cluster_name
            Name of the cluster
        rules
            List of rules to be looked for
        state
            How do you want the rules?
        """

        cluster = self._find_cluster_by_name(cluster_name)
        changed = False
        rules_set = set(rules)
        new_rules = []

        for rule in self.api.db_firewall_list(cluster.id):
            rc = DatabaseFirewallRuleCreate(type=rule.type, value=rule.value)

            if state == EntryState.present:
                new_rules.append(rule)

                if rc in rules_set:
                    rules_set.remove(rc)
            else:
                if rc not in rules_set:
                    new_rules.append(rule)
                else:
                    changed = True

        if rules_set and state == EntryState.present:
            new_rules.extend(rules_set)
            changed = True

        if changed:
            self.api.db_firewall_update(cluster.id, new_rules)

        return Outcome(changed)

    def db_firewall_droplets(
        self, cluster_name: Text, droplet_names: Sequence[Text]
    ) -> Outcome:
        """
        Allows a given list of droplets (based on their name) to access the
        cluster.

        Parameters
        ----------
        cluster_name
            Name of the cluster for which you're doing this
        droplet_names
            Names of the droplets you want to whitelist
        """

        cluster = self._find_cluster_by_name(cluster_name)
        droplet_names = set(droplet_names)
        droplet_ids = []

        for droplet in self.api.droplet_list():
            if droplet.name in droplet_names:
                droplet_ids.append(droplet.id)
                droplet_names.remove(droplet.name)

            if not droplet_names:
                break

        if droplet_names:
            raise IdemApiError(f"Droplets not found: {droplet_names!r}")

        return self.db_firewall_rule(
            cluster.name,
            [
                DatabaseFirewallRuleCreate(DbFirewallRuleType.droplet, f"{x}")
                for x in droplet_ids
            ],
            EntryState.present,
        )

    def psql_cluster(
        self,
        name: Text,
        region: Text,
        size: DatabaseSize,
        nodes: int,
        version: PostgreSqlVersion = DEFAULT_PSQL_VERSION,
        private_network: Optional[Uuid] = None,
        skip_checks: bool = False,
    ) -> Outcome:
        """
        Makes sure that this cluster exists. If the existing cluster doesn't
        match the specifications, this function will fail and not attempt to
        make any changes.

        Parameters
        ----------
        name
            Name of the cluster
        region
            DigitalOcean region for that cluster (by example: "ams3"), see the
            documentation to get those.
        size
            Size of the cluster
        nodes
            How many nodes do you want? Minimum 1, maximum 3 (except on the
            smallest size which can have only 1 node)
        version
            Version of PostgreSQL
        private_network
            ID of the private network to connect this cluster to. If not
            specified, it will get connected to the default network for this
            region.
        skip_checks
            Don't check that existing clusters match specifications
        """

        cluster = self._find_cluster_by_name(name)
        changed = False

        if cluster is None:
            changed = True
            cluster = self.api.db_cluster_create(
                DatabaseClusterCreate(
                    name=name,
                    engine=DatabaseEngine.pg,
                    version=version,
                    size=size,
                    region=region,
                    num_nodes=nodes,
                    private_network_uuid=private_network,
                )
            )

        if cluster.status != DatabaseStatus.online:
            start = time()

            for _ in range(0, self.PROVISION_TIMEOUT // self.PROVISION_POLL + 1):
                sleep(self.PROVISION_POLL)
                cluster = self.api.db_cluster_get(cluster.id)

                if cluster.status == DatabaseStatus.online:
                    break

                if time() - start > self.PROVISION_TIMEOUT:
                    break

        if cluster.status != DatabaseStatus.online:
            raise IdemApiError("Cluster failed to come online")

        if not skip_checks:
            if cluster.size != size:
                raise IdemApiError("Existing cluster does not have the right size")

            if cluster.region != region:
                raise IdemApiError("Existing cluster is not in the right region")

            if cluster.num_nodes != nodes:
                raise IdemApiError(
                    "Existing cluster does not have the right nodes number"
                )

        self._cluster_cache[cluster.name] = cluster

        return Outcome(changed)

    def psql_database(
        self,
        cluster_name: Text,
        name: Text,
        state: Text,
        copy_db_name: Text = "",
    ) -> Outcome:
        """
        Makes sure that this database exists. If the copy_db_name is specified,
        then the content will be copied from that other database (only at
        create time, not on subsequent runs).

        Notes
        -----
        Copy uses "pg_dump" and "psql", of the binaries are not installed on
        your system this will fail.

        Also, after the copy, the owner of tables will be doadmin. This is
        fixed when you attribute a user to this database using the psql_user()
        method.

        Parameters
        ----------
        cluster_name
            Name of the cluster to create the DB into
        name
            Name of the DB you want
        copy_db_name
            If creating the DB, copy the content from that other DB (from the
            same cluster)
        state
            Indicates the desired state for the database within the cluster: present or absent
        """

        changed = False
        cluster = self._find_cluster_by_name(cluster_name)

        absent = state == "absent"
        present = state == "present"

        if not cluster:
            raise IdemApiError(f"Cluster {cluster_name!r} does not exist")

        db = self._find_database_by_name(cluster.id, name)

        if not db and absent:
            return Outcome(False)

        with self._allow_self_access(cluster.name):
            if not db:
                changed = True
                self.api.db_database_create(cluster.id, Database(name=name))
            else:
                if absent:
                    changed = True
                    self.api.db_database_delete(cluster.id, database_name=name)

            if changed and copy_db_name and present:
                with NamedTemporaryFile() as f_info:
                    env = {**environ, **cluster.connection.pg_env}

                    with open(f_info.name, "wb") as f:
                        ret = subprocess.run(
                            [
                                "pg_dump",
                                "-O",
                                *cluster.connection.pg_flags(copy_db_name),
                            ],
                            env=env,
                            stdout=f,
                            stderr=subprocess.PIPE,
                        )

                        if ret.returncode:
                            raise IdemApiError(
                                f"Error while dumping DB: {ret.stderr[0:1000]}"
                            )

                    with open(f_info.name, "rb") as f:
                        ret = subprocess.run(
                            ["psql", *cluster.connection.pg_flags(name)],
                            env=env,
                            stdin=f,
                            capture_output=True,
                        )

                        if ret.returncode:
                            raise IdemApiError(
                                f"Error while restoring DB dump: {ret.stderr[0:1000]}"
                            )

        return Outcome(changed)

    def psql_user(
        self,
        cluster_name: Text,
        user_name: Text,
        db_name: Text,
        state: Text,
        pool_size: int = 1,
    ):
        """
        Ensures a user which has the rights to access the "db_name" database.
        If the user is being created, then all tables from "db_name" will be
        reassigned to this user. This is to fix the ownership of tables from
        the copy (see psql_database()).

        If a non-zero pool size is specified, a pool of that size will be
        created for this user.

        The outcome of this function will contain connection parameters for
        this user, both on the public and the private network. Please note
        that if a pool was created then the connection settings will be those
        of the pool instead of a direct connection.

        If present is set to False and the user exists, it will be deleted.

        Parameters
        ----------
        cluster_name
            Name of the cluster you want to create this user in
        user_name
            Name of the user you're creating
        db_name
            Name of the database for which this user is destined
        pool_size
            Size of the connection pool. Put 0 if you don't want any pool to
            be created.
        state
            Indicates the desired state for the user: present or absent
        """

        cluster = self._find_cluster_by_name(cluster_name)
        user = None
        changed = False

        absent = state == "absent"
        present = state == "present"

        for candidate in self.api.db_user_list(cluster.id):
            if candidate.name == user_name:
                user = candidate
                break

        if not user:
            if absent:
                return Outcome(False)

            changed = True
            user = self.api.db_user_create(
                cluster.id, DatabaseUserCreate(name=user_name)
            )

            with self._allow_self_access(cluster.name):
                ret = subprocess.run(
                    ["psql", *cluster.connection.pg_flags(db_name)],
                    env={**environ, **cluster.connection.pg_env},
                    capture_output=True,
                    input=f'reassign owned by "{cluster.connection.user}" to "{user_name}"'.encode(
                        "utf-8"
                    ),
                )

            if ret.returncode:
                raise IdemApiError(
                    f"Error while re-assigning DB tables to user: {ret.stderr[0:1000]}"
                )
        else:
            if absent:
                changed = True
                self.api.db_user_delete(cluster.id, user_name=user_name)

        pool = None
        effective_pool_name = f"user_{user_name}"

        for candidate in self.api.db_pool_list(cluster.id):
            if candidate.name == effective_pool_name:
                pool = candidate
                break

        if not pool and pool_size:
            if present:
                changed = True
                pool = self.api.db_pool_create(
                    cluster.id,
                    DatabaseConnectionPoolCreate(
                        name=effective_pool_name,
                        mode=PgBouncerMode.transaction,
                        size=pool_size,
                        db=db_name,
                        user=user_name,
                    ),
                )

        if pool and absent:
            changed = True
            self.api.db_pool_delete(cluster.id, effective_pool_name)
            pool = None

        if pool:
            private_connection = pool.private_connection
            connection = pool.connection
        else:
            private_connection = replace(
                cluster.private_connection,
                user=user.name,
                database=db_name,
                password=user.password,
            )
            connection = replace(
                cluster.connection,
                user=user.name,
                database=db_name,
                password=user.password,
            )

        return Outcome(
            changed,
            {
                "private_connection": private_connection,
                "connection": connection,
            },
        )



from collections import ChainMap
from dataclasses import Field, fields, is_dataclass
from typing import Any, Text, Union, get_type_hints

# noinspection PyPep8Naming
from uuid import UUID as BaseUuid

from typefit.compat import get_args, get_origin
from typefit.meta import Source
from typefit.serialize import SaneSerializer


class Uuid(BaseUuid):
    def __new__(cls, uuid: Text) -> "Uuid":
        self = super().__new__(cls)
        # noinspection PyTypeChecker
        cls.__init__(self, uuid)
        return self

    def __typefit_serialize__(self):
        return f"{self}"


def is_optional(field_type):
    """
    Checks fields that should be removed from output JSON if null, aka the
    fields that are mapped to a dataclass or a NamedTuple and which are
    optional. If they are None, we consider they shouldn't be in the output.
    """

    if not (get_origin(field_type) is Union):
        return False

    found_none = False
    found_dataclass = False

    for arg in get_args(field_type):
        if arg is not None and arg is not None.__class__:
            found_none = True
        elif is_dataclass(arg):
            found_dataclass = True
        elif issubclass(arg, tuple) and hasattr(arg, "_fields"):
            found_dataclass = True

    return found_dataclass and found_none


class UndefinedSerializer(SaneSerializer):
    """
    Removes from the output JSON the optional fields if they are not
    specified.

    See Also
    --------
    is_optional
    """

    def serialize_dataclass(self, obj: Any):
        """
        Same as parent but twisted a little bit. The "optionality test" happens
        before reverse source mapping is done, that's maybe not what you
        expected.
        """

        def _get_values():
            field: Field
            for field in fields(obj.__class__):
                if is_optional(field.type) and getattr(obj, field.name) is None:
                    continue

                if field.metadata and "typefit_source" in field.metadata:
                    source: Source = field.metadata["typefit_source"]
                    yield {
                        k: self.serialize(v)
                        for k, v in source.value_to_json(field.name, obj).items()
                    }
                else:
                    yield {field.name: self.serialize(getattr(obj, field.name))}

        return dict(ChainMap(*_get_values()))

    def serialize_tuple(self, obj: tuple):
        """
        Same as parent but with an additional check to test if the field is
        optional.
        """

        return {
            k: self.serialize(getattr(obj, k))
            for k, field_type in get_type_hints(obj.__class__).items()
            if not (is_optional(field_type) and getattr(obj, k) is None)
        }
